package eu.eiting.StarCharge.control;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import eu.eiting.StarCharge.control.controller.BaseController;
import eu.eiting.StarCharge.control.states.BaseState;
import eu.eiting.StarCharge.control.states.GameState;
import eu.eiting.StarCharge.control.states.MenuState;
import eu.eiting.StarCharge.control.states.PauseState;
import eu.eiting.StarCharge.control.states.SolarSystemState;
import eu.eiting.StarCharge.model.BattleField;
import eu.eiting.StarCharge.model.GameModel;
import eu.eiting.StarCharge.model.listener.ChangeListener;
import eu.eiting.StarCharge.visualization.GameFrame;

/**
 * The main interface to the game. 
 * 
 *   - contains methods to change the state of the game
 *   - contains methods to start or load a game
 *    
 * @author Caspar Eiting
 *
 */
public class StarCharge extends Thread{
	// CLASS ATTRIBUTES
	//====================================
	public static final boolean DEBUG													= true;
	public static final long FRAMES_PER_SECOND											= 40;
	private static StarCharge instance													= null;
	private static final Set<ChangeListener<GameModel>> gameModelChangeListeners		= new HashSet<>();
	private static final Set<ChangeListener<BaseState.States>> gameStateChangeListeners	= new HashSet<>();
	//====================================
	
	// CLASS METHODS
	//====================================
	public static void addGameModelListener(ChangeListener<GameModel> listener){
		gameModelChangeListeners.add(listener);
	}
	
	public static void removeGameModelListener(ChangeListener<GameModel> listener){
		gameModelChangeListeners.remove(listener);
	}
	
	public static void addGameStateListener(ChangeListener<BaseState.States> listener){
		gameStateChangeListeners.add(listener);
	}
	
	public static void removeGameStateListener(ChangeListener<BaseState.States> listener){
		gameStateChangeListeners.remove(listener);
	}
	
	public static void changeState(BaseState.States state){
		BaseState nextState = instance.states.get(state);
		
		if(nextState != null) {
			instance.currentState = nextState;

			instance.currentStateType = state;

			for (ChangeListener<BaseState.States> listener : gameStateChangeListeners) {
				listener.onChange(state);
			}
		}
	}
	
	public static void loadGame(String path){
		instance.setGameModel(GameModel.Load(path));
	}
	
	public static void startGame(){
		instance.setGameModel(GameModel.New());
	}
	
	public static BaseState.States getCurrentStateType(){
		return instance.currentStateType;
	}
	
	// TODO: move to own Class (Collision/Physic)
	public static BattleField getCurrentBattlefield(){
		return instance.model.getBattleField();
	}
	//====================================
	
	// OBJECT ATTRIBUTES
	//====================================
	private BaseState currentState							= null;
	private BaseState.States currentStateType;
	private final Map<BaseState.States, BaseState> states	= new HashMap<BaseState.States, BaseState>();

	private GameFrame gameFrame								= null;
	
	private GameModel model									= null;
	
	private boolean isRunning								= true;
	private Timer frameTimer								= new Timer(1.0f / FRAMES_PER_SECOND);
	//====================================
	
	// CONSTRUCTOR
	//====================================
	public StarCharge(){
		instance = this;

		gameFrame = new GameFrame();
		
		// PREPARE STATES
		//====================================
		states.put(BaseState.States.MENU_STATE, new MenuState());
		states.put(BaseState.States.GAME_STATE, new GameState());
		states.put(BaseState.States.PAUSE_STATE, new PauseState());
		states.put(BaseState.States.SOLAR_SYSTEM_STATE, new SolarSystemState());
		//====================================
				
		// SET INITIAL STATE
		//====================================
		currentState = states.get(BaseState.States.MENU_STATE);
		//====================================
	}
	//====================================
	
	// OBJECT METHODS
	//====================================
	/**
	 * This method controls the execution of the mechanics 
	 * and the visualization of the game. It contains the
	 * while-loop that represents the GameLoop.
	 */
	@Override
	public void run(){
		while (isRunning) {
			frameTimer.start();

			for (BaseController selected : gameFrame.getController()) {
				if(selected != null){
					selected.update();
				}
			}
			
			currentState.execute();

			gameFrame.repaint();
			
			try {
				Thread.sleep(frameTimer.getRemainingTime());
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}
	//====================================
	
	// GETTER AND SETTER
	//====================================
	private void setGameModel(GameModel model){
		this.model = model;
		
		for (ChangeListener<GameModel> listener : gameModelChangeListeners) {
			listener.onChange(model);
		}
	}
	//====================================
}
