package eu.eiting.StarCharge.model.structure;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

/**
 * A class that represents a position, rotation and direction in world space and local space.
 * 
 *  - contains methods rotate and transform the data
 *  - can be parented to another transform
 * 
 * @author Caspar Eiting
 *
 */
public class Transform implements Serializable{
	// CLASS ATTRIBUTES
	// ============================================
	private static final long serialVersionUID = -1551599839670746095L;
	// ============================================
	
	// OBJECT ATTRIBUTES
	// ============================================
	private Vector2 position				= Vector2.ZERO;
	private Vector2 localPosition			= Vector2.ZERO;
	private Vector2 direction				= Vector2.FORWARD;
	private Vector2 localDirection			= Vector2.FORWARD;
	private double rotation					= 0.0;
	private double localRotation			= 0.0;
	private Transform parent				= null;
	private final List<Transform> CHILDREN	= new ArrayList<>();
	// ============================================
	
	// OBJECT METHODS
	// ============================================
	public void rotate(double degree){
		setRotation(this.localRotation + degree);		
	}
	
	private double trimRotation(double value){
		if(value > 359){
			return value % 360;
		} else if(value < 0){
			return 360 + value % 360;
		} else {
			return value;
		}
	}
	// ============================================
	
	// GETTER AND SETTER
	// ============================================
	public double getRotation() {
		return rotation;
	}

	public void setRotation(double degree) {
		if(parent != null){
			this.localRotation	= trimRotation(degree);
			this.rotation		= trimRotation(parent.getRotation() + degree);
			
			this.localDirection = Vector2.rotate(Vector2.FORWARD, Math.toRadians(localRotation));
			this.direction		= Vector2.rotate(Vector2.FORWARD, Math.toRadians(rotation));

			setPosition(localPosition);
		} else {
			this.localRotation	= rotation = trimRotation(degree);
			this.localDirection = direction = Vector2.rotate(Vector2.FORWARD, Math.toRadians(rotation));
		}

		for (Transform child : CHILDREN) {
			child.setRotation(child.getLocalRotation());
		}
	}
	
	public void setDirection(Vector2 direction) {
		setRotation(Vector2.getAngle(Vector2.FORWARD, direction));
	}
	
	public void setParent(Transform parent) {
		if(parent != this){
			if(this.parent != null){
				this.parent.CHILDREN.remove(this);
			}

			this.parent = parent;
		
			if(this.parent != null){
				this.parent.CHILDREN.add(this);
			
				setPosition(getPosition());
			}
		} else {
			throw new IllegalArgumentException("It is not possible to parent a Transform to itself");
		}
	}
	
	public double getLocalRotation() {
		return localRotation;
	}

	public Vector2 getLocalPosition() {
		if(parent != null){
			return localPosition;
		}
		
		return position;
	}
	
	public Vector2 getPosition() {
		return position;
	}
	
	public void setPosition(Vector2 position){
		if(parent == null){
			this.position		= localPosition = position;
		} else {
			this.localPosition	= position;
			this.position		= Vector2.add(parent.getPosition(), Vector2.rotate(localPosition, Math.toRadians(parent.getLocalRotation())));
		}
		
		for (Transform child : CHILDREN) {
			child.setPosition(child.getLocalPosition());
		}
	}

	public Vector2 getDirection() {
		return direction;
	}
	
	public Vector2 getLocalDirection() {
		return localDirection;
	}
	// ============================================
}
