package eu.eiting.StarCharge.model.structure;

import java.io.Serializable;

public class Vector2 implements Serializable{
	// CLASS ATTRIBUTES
	//====================================
	private static final long serialVersionUID	= -3306176728794796403L;
	public static final Vector2 ZERO			= new Vector2(0, 0);
	public static final Vector2 FORWARD			= new Vector2(0,  -1);
	public static final Vector2 BACKWARD		= new Vector2(0,  1);
	public static final Vector2 LEFT			= new Vector2(-1,  0);
	public static final Vector2 RIGHT			= new Vector2(1,  0);
	//====================================
	
	// OBJECT ATTRIBUTES
	//====================================
	public final float x;
	public final float y;
	public final float magnitude;
	//====================================
	
	// CONSTRUCTOR
	//====================================
	public Vector2(float x, float y){
		this.x			= x;
		this.y			= y;
		this.magnitude	= length(x, y);
	}
	
	public Vector2(float x, float y, float length){
		this( ((x / length(x, y)) * length), ((y / length(x, y)) * length));
	}
	
	public Vector2(Vector2 v, float length){
		this( ((v.x / length(v.x, v.y)) * length), ((v.y / length(v.x, v.y)) * length));
	}
	//====================================

	// OBJECT METHODS
	//====================================
	@Override
	public String toString() {
		return "Vector2 [x=" + x + ", y=" + y + "]";
	}
	
	@Override
	public int hashCode() {
		return (int) (x + y);
	}
	
	@Override
	public boolean equals(Object obj) {
		if (obj instanceof Vector2) {
			if (obj.hashCode() == this.hashCode()) {
				Vector2 other = (Vector2) obj;
				
				if (x == other.x) {
					if (y == other.y) {
						return true;
					}
				}
			}
		}
		
		return false;
	}
	//====================================
	
	// CLASS METHODS
	//====================================
	public static Vector2 add(Vector2 a, Vector2 b){
		return new Vector2(a.x + b.x, a.y + b.y);
	}
	
	public static Vector2 substract(Vector2 a, Vector2 b){
		return new Vector2(a.x - b.x, a.y - b.y);
	}
	
	public static Vector2 multiply(Vector2 a, float value){
		return new Vector2(a.x * value, a.y * value);
	}
	
	public static Vector2 devide(Vector2 a, float value){
		return new Vector2(a.x / value, a.y / value);
	}
	
	public static double dotProduct(Vector2 a, Vector2 b){
		return (a.x * b.x) + (a.y * b.y);
	}
	
	public static float length(Vector2 a){
		return (float) Math.sqrt(dotProduct(a, a));
	}
	
	public static float length(float x, float y){
		return (float) Math.sqrt((x * x) + (y * y));
	}
	
	public static float distance(Vector2 a, Vector2 b){
		return length(substract(a, b));
	}
	
	public static Vector2 normalize(Vector2 a){
		float length =  length(a);
		
		if(length != 0){
			return devide(a, length);
		} else {
			return ZERO;
		}
	}
	
	public static Vector2 createWithLength(Vector2 v, float length){
		return Vector2.multiply(Vector2.normalize(v), length);
	}
	
	public static Vector2 left(Vector2 a){
		return new Vector2(-a.y, a.x);
	}
	
	public static Vector2 right(Vector2 a){
		return new Vector2(a.y, -a.x);
	}
	
	public static Vector2 invert(Vector2 v){
		return new Vector2(-v.x, -v.y);
	}
	
	/**
	 * Calculates angle between 0.0 and 360.0 degrees over right
	 * @param a
	 * @param b
	 * @return
	 */
	public static double getAngle(Vector2 a, Vector2 b){
		double angle =Math.atan2(b.x, b.y) - Math.atan2(a.x, a.y);
		
		if(angle > Math.PI){
			angle -= Math.PI * 2;
		} else if(angle < -Math.PI){
			angle += Math.PI * 2;
		}

		return -Math.toDegrees(angle);
	}
	
	public static double getRadians(Vector2 a, Vector2 b){
		double radian =Math.atan2(b.x, b.y) - Math.atan2(a.x, a.y);
		
		if(radian > Math.PI){
			radian -= Math.PI * 2;
		} else if(radian < -Math.PI){
			radian += Math.PI * 2;
		}

		return radian;
	}

	public static Vector2 rotate(Vector2 v, double radian){
		// ROTATION MATRIX
		//====================================
		float x = (float) (Math.cos(radian) * v.x - Math.sin(radian) * v.y);
		float y = (float) (Math.sin(radian) * v.x + Math.cos(radian) * v.y);
		//====================================
		
		// TODO: IMPROVE - temporary solution 
		// 				   to remove precision error
		//====================================
		if(x == 1.0f || x == -1.0f){
			y = 0.0f;
		} else if(y == 1.0f || y == -1.0f){
			x = 0.0f;
		}
		//====================================

		return new Vector2(x, y);
	}
	//====================================
}
